# -*- coding: utf-8 -*-

import time
from urlparse import urlparse

from newspider.comm.utils import Utils


class Guarder:

    def __init__(self,config):
        self.config = config
        self.map_last_req = {}

    def check(self,url):
        interval = self.config.get('GUARD_INTERVAL',0)
        if interval == 0: return True # 0表示禁用门卫功能,所以直接返回即可

        host = urlparse(url).hostname
        current_time = int(time.time())
        last_time = self.map_last_req.get(Utils.md5(host),0)
        if current_time - last_time > interval:
            self.map_last_req[Utils.md5(host)] = current_time #记录这次的时间
            return True
        else:
            return False


# -*- coding: utf-8 -*-
import os
from urlparse import urlparse

from newspider.comm.utils import Utils


class UrlRecorder:
    CACHE_DIR = '.cache'

    def __init__(self,fsync=True):

        if not os.path.exists(UrlRecorder.CACHE_DIR):
            os.makedirs(UrlRecorder.CACHE_DIR)

        self.map_files = {}
        self.map_cache = {}
        self.fsync = fsync

    def record(self,url):
        # 如果命中了,直接返回即可,无需多吃记录
        if self.hit(url): return

        objUrl = urlparse(url)
        if objUrl is None: return

        host = objUrl.hostname

        if self.fsync: #如果不需要持久保存,则不写文件
            cache_file = "%s/%s.lst" % (UrlRecorder.CACHE_DIR, host)

            file = self.map_files.get(cache_file,None)
            if not file or  file.closed :
                file =  open(cache_file,'a')
                self.map_files[cache_file] = file

            file.write("%s\n" % Utils.md5(url))

        # 游湖文件没有那么快刷磁盘,这里需要在缓存里面多刷一次,免得下次过来可能取不到
        tmp = self.map_cache.get(host,[])
        tmp.append(Utils.md5(url))
        self.map_cache[host] = tmp

    def hit(self,url):
        objUrl = urlparse(url)
        if objUrl is None: return

        host = objUrl.hostname
        list = self.map_cache.get(host,None)

        if list is None and self.fsync:
            # 尝试从文件中加载,这里不考虑内存过大的情况
            cache_file = "%s/%s.lst" % (UrlRecorder.CACHE_DIR, host)
            if os.path.exists(cache_file):
                list = map(lambda x: x.strip(),open(cache_file,'r').readlines())
                self.map_cache[host] = list
        # end if

        md5 = Utils.md5(url)
        return True if list is not None and md5 in list else False

    def shutdown(self):
        for file in self.map_files.keys():
            f = self.map_files.get(file)
            if file is not None: f.close()


if __name__ == '__main__':

    url = "http://www.baidu.com/aaaa?kkk=aaa&bbb=sss"
    obj = UrlRecorder()
    # assert  obj.hit(url) == False

    obj.record(url)
    assert obj.hit(url) == True

    url2 = "http://www.baidu.com/aaaa?kkk=测试中文"

    obj.record(url2)
    assert obj.hit(url2) == True


    obj.shutdown()




# -*- coding: utf-8 -*-
import requests
import logging

class Req:

    def __init__(self,config):
        self.config = config
        self.session = requests.session()
        self.logger = logging.getLogger('Newspider')


    def get(self,url):
        user_agent = self.config.get('USER_AGENT','newspider/1.0')
        headers = {'user-agent': user_agent}

        #检测代理
        http_proxy = self.config.get('HTTP_PROXY',None)
        proxies = {}
        if http_proxy is not None:
            proxies['http'] = http_proxy

        self.logger.debug("Requesting content for url: %s" % url)
        r = self.session.get(url, headers=headers,proxies = proxies)
        self.logger.debug("Response code: %d" % r.status_code)
        if r.content is not None:
            return r.content
        else:
            return None





# -*- coding: utf-8 -*-
class Utils:

    @classmethod
    def tag_of_object(cls, obj):
        tag = str(obj.__class__).split('.')[-1]
        return tag

    @classmethod
    def md5(cls,str):
        import md5
        m1 = md5.new()
        m1.update(str)
        return m1.hexdigest()

    @classmethod
    def get_attr_in_class(cls,className, attr,default):
        try:
            res = eval("%s.%s" %(className, attr))
            return res
        except Exception,e:
            return default

    @classmethod
    def conv2utf8(cls,content,from_encodeing="GBK"):
        return content.decode(from_encodeing,errors='ingore')#.encode('utf-8')



# -*- coding: utf-8 -*-
# import sys
# reload(sys)
# sys.setdefaultencoding("utf-8")

from pyquery import PyQuery as pq
import simplejson as JSON
import re
import time

from newspider.interfaces import *
from newspider.spider import Newspider
from newspider.comm.utils import Utils
# from base.commit import PostSession
# from config import *



'''
微博抓取器
'''

class WeiboFetcher(IntFetcher):

    def __init__(self):
        self.next_page = []

    def fetch_detail_urls(self,html):
        #print "aaaa"

        htmlContent = Utils.conv2utf8(html,'GBK')

        #print "bbb"
        #print html
        tag = u"全部"
        l = []

        #htmlContent = unicode(html,'GBK')
        html = pq(htmlContent)
        array = html(".list14 ul li a")
        i = -1
        for item in array:
            element = pq(item)
            i = i + 1
            if i == 0:
                tag = element.text()
                continue
            if i % 2 == 0:
                continue
            url = element.attr('href')
            # text = element.text()
            l.append((url,{"category":tag}))
        return l

    def start_page(self):
        return [  "http://roll.sohu.com/sports_bak/",#>体育</a></li>
            # "http://roll.sohu.com/yule_bak/",#>娱乐</a></li>
            # "http://roll.sohu.com/it_bak/" ,#IT数码</a></li>
            # "http://roll.sohu.com/auto_bak/",# >汽车</a></li>
            # "http://roll.sohu.com/focus_bak/",#>房产</a></li>
            # "http://roll.sohu.com/women_bak/",#>女性</a></li>
            # "http://roll.sohu.com/fashion_bak/",#>时尚</a></li>
            # #"http://roll.sohu.com/baobao/",# >母婴</a></li>-->
            # "http://roll.sohu.com/health_bak/",# >健康</a></li>
            # "http://roll.sohu.com/cul_bak/",# >文化</a></li>
            # "http://roll.sohu.com/learning_bak/",# >教育</a></li>
            # "http://roll.sohu.com/money_bak/",# >财经</a></li>
            # "http://roll.sohu.com/stock_bak/",# >证券</a></li>
            # "http://roll.sohu.com/games_bak/",# >游戏</a></li>
            # "http://roll.sohu.com/media_bak/",# >媒体</a></li>
            # #"http://roll.sohu.com/city_bak/",# >城市</a></li>
            # #"http://roll.sohu.com/luxury_bak/",# >奢侈品</a></li>
            # #"http://roll.sohu.com/star_bak/",# >评论</a></li>
            # "http://roll.sohu.com/travel/",# >旅游</a></li>
            # #"http://roll.sohu.com/pic/",# >大视野</a></li>
            # #"http://roll.sohu.com/subject_bak/",# >专题</a></li>
          ]

    def next_pages(self):
        time.sleep(1)
        return "http://roll.sohu.com/sports_bak/?t=%d" % int(time.time())

    def getJson(self, temp):
        json_str = temp.replace("FM.view(", "").replace(")</script>", "")
        # print json_str
        json = JSON.loads(json_str)
        return json


class WeiboParser(IntParser):

    def parse(self,tag,htmlContent,extras):
        pass


    def filterUrl(self, url):
        url.replace(" ", "")
        p = re.compile(r"http[s]?://(?:[a-zA-Z]|[0-9]|[-.]|_|/|(?:%[0-9a-fA-F][0-9a-fA-F]))+")
        match = p.findall(url)
        if match:
            print "match %s" % match[0]
            return match[0]
        else:
            return url

    def filterContentText(self, contentText):
        return contentText.strip().replace("\n", '').replace('\t', '')


def test_ctr_c(sp):
    time.sleep(1)
    sp.sigint_handler(None,None)



if __name__ == '__main__':
    sp = Newspider()
    sp.config("USER_AGENT","Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html")
    sp.config('GUARD_INTERVAL', 0)
    sp.config('DEBUG', True)


    sp.add_parser(WeiboParser())
    sp.add_fetcher(WeiboFetcher())

    # import threading
    # t = threading.Thread(target=test_ctr_c, args=(sp,))
    # t.start()

    sp.run()






# -*- coding: utf-8 -*-
import Queue
import logging
import signal
import time
import traceback

import threadpool
from comm.guarder import Guarder
from comm.urlrecoder import UrlRecorder
from comm.utils import Utils

from newspider.comm.req import Req
from newspider.interfaces import *


class Newspider:
    TASK_DONE_FLAG = "TASK_DONE"
    TASK_SHUTDOWN_FLAG = 'TASK_SHUTDOWN'

    def __init__(self):
        signal.signal(signal.SIGINT, self.sigint_handler)
        self.configuration = {}
        self.init_default_config()
        self.logger = logging.getLogger('Newspider')

        self.fetchers = []
        self.parsers = []
        self.running = True

        self.workers = threadpool.ThreadPool(self.configuration.get('WORKER_NUM', 5))
        self.category_urls = Queue.Queue(maxsize = 100)
        self.details_urls = Queue.Queue(maxsize = 100 )
        # self.fetchers_map = {}

        self.req = Req(self.configuration)
        self.guarder = Guarder(self.configuration)
        self.recorder = UrlRecorder(True)
        self.category_url_record = UrlRecorder(False)

    def init_default_config(self):
        self.configuration['USER_AGENT']= 'Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/48.0.2564.23 Mobile Safari/537.36'

        logger = logging.getLogger('Newspider')
        logger.setLevel(logging.DEBUG)
        hdr = logging.StreamHandler()
        # formatter = logging.Formatter('[%(asctime)s] %(name)s:%(levelname)s: %(message)s')
        # hdr.setFormatter(formatter)
        logger.addHandler(hdr)

    def config(self,key,value):
        """
        配置参数
        :param key: 可用值： USER_AGENT| WORKER_NUM| GUARD_INTERVAL
        :param value:
        :return:
        """
        self.configuration[key] = value


    def run(self):
        #特殊处理一下
        if self.is_debug():
            self.recorder = UrlRecorder(False) #重新生成一个不写文件的记录器,方便调试

        self.running = True
        self.logger.info("Start to run Newspider with %d fetcher(s) and %d parser(s)..." %(len(self.fetchers),len(self.parsers)))

        #第一步,先将start_page丢到队列中
        for fet in self.fetchers:
            if hasattr(fet,'start_page'):
                urls = fet.start_page()
                if type(urls) is list: [self.category_urls.put((fet, url)) for url in urls]
                else:self.category_urls.put((fet, urls))

        reqs = threadpool.makeRequests(self.do_fetch_details_urls, (self.category_urls,))
        [self.workers.putRequest(req) for req in reqs]

        reqs = threadpool.makeRequests(self.do_fetch_details_content, [self.details_urls] * int(self.configuration.get('WORKER_NUM', 5) - 1))
        [self.workers.putRequest(req) for req in reqs]

        self.workers.wait()

    def do_fetch_details_urls(self,queue):
        """
        【注意】只能在一个线程中处理此函数，不能并行
        处理分类页面，并在分类页面里面抽取子页面也next_page
        :param queue:
        :return:
        """
        self.logger.info("Starting worker for fetching details urls in category page...")
        while self.running and not queue.empty():
            """
            处理逻辑：
            2. 检查门卫系统，如果未满足条件，则将url丢到队列尾部，否则继续
            3. 请求html，
            4. 解析details_url列表，并添加到url列表队列
            5. 解析next_page_url，并添加到queue队列尾部
            """
            try:
                data = queue.get()
                fet, url = data
                if fet is None or url is None: continue

                # 门卫系统检查,如果未到时间,则放回队列中继续等待
                if not self.guarder.check(url):
                    self.logger.debug("Guarder check failed, skip this url for this time")
                    self.category_urls.put((fet, url))
                    time.sleep(1)
                    continue

                # 首先拉取html内容
                html = self.req.get(url)
                if html is not None: self.logger.debug("Got html content from %s" % url)

                # 解析当前页面中的所有详情页面url
                if not hasattr(fet, 'fetch_detail_urls'):
                    self.logger.error("Not method name fetch_detail_urls in fetcher %s" % fet.__class__)
                    continue

                try:
                    url_lists = fet.fetch_detail_urls(html)
                except Exception, e:
                    self.logger.error("Error when calling fetcher's fetch_detail_urls(): %s" % e.message)
                    traceback.print_exc()
                    url_lists=[]

                tag = Utils.tag_of_object(fet)
                self.logger.debug("Found %d detail urls for tag %s. " % (len(url_lists), tag))
                for item in url_lists:
                    url,extras = item
                    self.details_urls.put((tag, url, extras))


                # 解析当前页面中所有下一页的url
                if not hasattr(fet, 'next_pages'):
                    self.logger.error("Not method name next_pages in fetcher %s" % fet.__class__)
                    break

                try:
                    next_page_urls = fet.next_pages()
                except Exception, e:
                    traceback.print_exc()
                    self.logger.error("Error when calling fetcher's fetch_detail_urls(): %s" % e.message)

                if type(next_page_urls) is list:
                    self.logger.debug("Found %d url(s) for next page" % len(next_page_urls))
                    for url in next_page_urls:
                        if not self.category_url_record.hit(url):
                            self.category_urls.put((fet, url))
                            self.category_url_record.record(url)
                else:
                    self.logger.debug("Found one url for next page.")
                    if not self.category_url_record.hit(url):
                        self.category_urls.put((fet, next_page_urls))
                        self.category_url_record.record(url)

            except Exception,e:
                self.logger.warn("Error when process one item from queue, skip it")
                traceback.print_exc()
                continue
        self.details_urls.put(Newspider.TASK_DONE_FLAG)
        self.logger.info("Category fetcher exist now")

    def do_fetch_details_content(self, queue):
        """
        消费者线程处理函数。处理details_urls列表中的url，并丢给parsers处理
        :param queue:
        :return:
        """
        while self.running:
            """
            处理逻辑：
            1. 如果遇到中止标记，并且队列为空，则中止此线程；否则将中止标记继续丢给队尾
            2. 检查门卫系统，如果检查不通过，则将data丢到队尾，中止当前过程
            3. 最后再丢给parser处理
            """
            try:
                data = queue.get(True, 2)
                if isinstance(data, basestring) and data == Newspider.TASK_SHUTDOWN_FLAG:
                    self.logger.info("Found shutdown flag from queue")
                    with queue.mutex: queue.queue.clear()
                    queue.put(data)
                    break

                if isinstance(data, basestring) and data == Newspider.TASK_DONE_FLAG:
                    if queue.empty():
                        self.logger.info("Found quit flag from queue.")
                        queue.put(data)
                        break
                    else:
                        queue.put(data)
                        continue

                if type(data) is tuple:
                    tag,url,extras = data
                    if self.recorder.hit(url):
                        self.logger.debug("Url has beed fetched, skip it. %s" % url)
                    else:
                        # 门卫系统检查,如果未到时间,则放回队列中继续等待
                        if not self.guarder.check(url):
                            self.logger.debug("Guarder check failed, skip this url for this time")
                            queue.put((tag, url, extras))
                            time.sleep(1)
                            continue

                        html = self.req.get(url)
                        if html is not None:
                            extras['_url'] = url
                            self.recorder.record(url)  # 继续扒取过的url
                            for p in self.parsers:
                                try:
                                    p.parse(tag, html, extras)
                                except Exception, e:
                                    traceback.print_exc()
                                    self.logger.error("Failed to parse html for url: %s" % url)


            except Exception, e: pass
        self.logger.info("Do_fetch_details_content worker exist now")

    def add_fetcher(self,fet):
        if isinstance(fet,IntFetcher):
            self.fetchers.append(fet)
        else:
            self.logger.warn("object add to fetchers must be subclass of IntFetcher")

    def add_parser(self,parser):
        if isinstance(parser,IntParser):
            self.parsers.append(parser)
        else:
            self.logger.warn("object add to parsers must be subclass of IntParser")

    def sigint_handler(self, signum, frame):
        self.logger.info('Catched interrupt signal, Newspider will exits now!')
        self.running = False
        self.details_urls.put(Newspider.TASK_SHUTDOWN_FLAG)
        #self.workers.dismissedWorkers(self.configuration.get('WORKER_NUM', 5), False)

    def is_debug(self):
        return self.configuration.get('DEBUG',False) is True

    def __del__(self):
        self.workers.dismissedWorkers(self.configuration.get('WORKER_NUM', 5), False)
        self.recorder.shutdown()





# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

setup(
    name="newspider",
    version="0.9.9",
    description="分类扒数据的简易框架",
    long_description="",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
    keywords=(""),
    author="Ray",
    author_email="miuyin@126.com",
    url="http://www.typechodev.com",
    license="MIT",
    packages=['newspider','newspider.comm'],
    include_package_data=True,
    zip_safe=True,
    install_requires=['threadpool', 'pyquery', 'requests'],
    platforms="any"
)



# -*- coding: utf-8 -*-
import threadpool
import Queue
import  time
import random
from threading import  current_thread

class Demo:
    def __init__(self):
        self.q = Queue.Queue()
        self.running = True

    def f1(self,queue):
        cnt = 0
        while self.running:
            time.sleep(random.randint(0, 2))
            queue.put(cnt)
            cnt += 1
            if cnt == 20:
                queue.put(-1)
                break

    def f2(self,queue):
        while self.running:
            time.sleep(random.randint(0, 3))
            try:
                data = queue.get(True,2)
                if data is not None:
                    if data == -1:
                        print "Found quit flag"
                        queue.put(data)
                        break
                    print "[%s] consumer pick %d" % (current_thread().name, int(data))
            except Exception,e: pass

    def run(self):
        pool = threadpool.ThreadPool(3)

        requests = threadpool.makeRequests(self.f1,(self.q,))
        [pool.putRequest(req) for req in requests]

        requests = threadpool.makeRequests(self.f2,(self.q,self.q,self.q))
        [pool.putRequest(req) for req in requests]
        pool.wait()

if __name__ == '__main__':
    d = Demo()
    d.run()


# -*- coding: utf-8 -*-


class IntFetcher:

    def fetch_detail_urls(self, content):
        """
        从content中解析二级页的url地址
        :param content: 一级页的html内容
        :return: 返回(list,extras)
        """
        raise Exception("Method fetch_detail_urls not supported yet")

    def start_page(self):
        """
        初始url
        :return: 返回list或者url
        """
        raise Exception("Method net_url not supported yet")

    def next_pages(self):
        """
        从当前一级页中解析next_page的url地址
        :return: list 或者 string
        """
        raise Exception("Method net_url not supported yet")


class IntParser:
    def parse(self,tag, content, url):
        raise Exception("Method parse not supported yet")



